@@ -10,14 +10,17 @@ module Data.Conduit.Binary
       -- ** Sources
       sourceFile
     , sourceHandle
+    , sourceHandleUnsafe
     , sourceIOHandle
     , sourceFileRange
+    , sourceHandleRange
       -- ** Sinks
     , sinkFile
     , sinkHandle
     , sinkIOHandle
       -- ** Conduits
     , conduitFile
+    , conduitHandle
       -- * Utilities
       -- ** Sources
     , sourceLbs
@@ -28,30 +31,39 @@ module Data.Conduit.Binary
     , drop
     , sinkCacheLength
     , sinkLbs
+    , mapM_
       -- ** Conduits
     , isolate
     , takeWhile
     , Data.Conduit.Binary.lines
     ) where
 
-import Prelude hiding (head, take, drop, takeWhile, dropWhile)
+import Prelude hiding (head, take, drop, takeWhile, dropWhile, mapM_)
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as L
 import Data.Conduit
 import Data.Conduit.List (sourceList, consume)
 import Control.Exception (assert, finally)
-import Control.Monad (unless)
+import Control.Monad (unless, when)
 import Control.Monad.IO.Class (liftIO, MonadIO)
 import Control.Monad.Trans.Resource (allocate, release)
+import Control.Monad.Trans.Class (lift)
 import qualified System.IO as IO
 import Data.Word (Word8, Word64)
 import Control.Applicative ((<$>))
 import System.Directory (getTemporaryDirectory, removeFile)
+import Data.ByteString.Lazy.Internal (defaultChunkSize)
 #if CABAL_OS_WINDOWS
 import qualified System.Win32File as F
 #elif NO_HANDLES
 import qualified System.PosixFile as F
 #endif
+import Data.ByteString.Internal (ByteString (PS), inlinePerformIO)
+import Foreign.ForeignPtr.Unsafe (unsafeForeignPtrToPtr)
+import Foreign.ForeignPtr (touchForeignPtr)
+import Foreign.Ptr (plusPtr)
+import Foreign.Storable (peek)
+import GHC.ForeignPtr           (mallocPlainForeignPtrBytes)
 
 -- | Stream the contents of a file as binary data.
 --
@@ -83,11 +95,36 @@ sourceHandle h =
     loop
   where
     loop = do
-        bs <- liftIO (S.hGetSome h 4096)
+        bs <- liftIO (S.hGetSome h defaultChunkSize)
         if S.null bs
             then return ()
             else yield bs >> loop
 
+-- | Same as @sourceHandle@, but instead of allocating a new buffer for each
+-- incoming chunk of data, reuses the same buffer. Therefore, the @ByteString@s
+-- yielded by this function are not referentially transparent between two
+-- different @yield@s.
+--
+-- This function will be slightly more efficient than @sourceHandle@ by
+-- avoiding allocations and reducing garbage collections, but should only be
+-- used if you can guarantee that you do not reuse a @ByteString@ (or any slice
+-- thereof) between two calls to @await@.
+--
+-- Since 1.0.12
+sourceHandleUnsafe :: MonadIO m => IO.Handle -> Source m ByteString
+sourceHandleUnsafe handle = do
+    fptr <- liftIO $ mallocPlainForeignPtrBytes defaultChunkSize
+    let ptr = unsafeForeignPtrToPtr fptr
+        loop = do
+            count <- liftIO $ IO.hGetBuf handle ptr defaultChunkSize
+            when (count > 0) $ do
+                yield (PS fptr 0 count)
+                loop
+
+    loop
+
+    liftIO $ touchForeignPtr fptr
+
 -- | An alternative to 'sourceHandle'.
 -- Instead of taking a pre-opened 'IO.Handle', it takes an action that opens
 -- a 'IO.Handle' (in read mode), so that it can open it only when needed
@@ -131,25 +168,34 @@ sourceFileRange :: MonadResource m
 sourceFileRange fp offset count = bracketP
     (IO.openBinaryFile fp IO.ReadMode)
     IO.hClose
-    start
-  where
-    start handle = do
-        case offset of
-            Nothing -> return ()
-            Just off -> liftIO $ IO.hSeek handle IO.AbsoluteSeek off
-        case count of
-            Nothing -> pullUnlimited handle
-            Just c -> pullLimited (fromInteger c) handle
+    (\h -> sourceHandleRange h offset count)
 
-    pullUnlimited handle = do
+-- | Stream the contents of a handle as binary data, starting from a certain
+-- offset and only consuming up to a certain number of bytes.
+--
+-- Since 1.0.8
+sourceHandleRange :: MonadIO m
+                  => IO.Handle
+                  -> Maybe Integer -- ^ Offset
+                  -> Maybe Integer -- ^ Maximum count
+                  -> Producer m S.ByteString
+sourceHandleRange handle offset count = do
+    case offset of
+        Nothing -> return ()
+        Just off -> liftIO $ IO.hSeek handle IO.AbsoluteSeek off
+    case count of
+        Nothing -> pullUnlimited
+        Just c -> pullLimited (fromInteger c)
+  where
+    pullUnlimited = do
         bs <- liftIO $ S.hGetSome handle 4096
         if S.null bs
             then return ()
             else do
                 yield bs
-                pullUnlimited handle
+                pullUnlimited
 
-    pullLimited c handle = do
+    pullLimited c = do
         bs <- liftIO $ S.hGetSome handle (min c 4096)
         let c' = c - S.length bs
         assert (c' >= 0) $
@@ -157,7 +203,7 @@ sourceFileRange fp offset count = bracketP
                 then return ()
                 else do
                     yield bs
-                    pullLimited c' handle
+                    pullLimited c'
 
 -- | Stream all incoming data to the given file.
 --
@@ -187,9 +233,15 @@ conduitFile :: MonadResource m
 conduitFile fp = bracketP
     (IO.openBinaryFile fp IO.WriteMode)
     IO.hClose
-    go
-  where
-    go h = awaitForever $ \bs -> liftIO (S.hPut h bs) >> yield bs
+    conduitHandle
+
+-- | Stream the contents of the input to a @Handle@, and also send it along the
+-- pipeline. Similar in concept to the Unix command @tee@. Like @sourceHandle@,
+-- does not close the handle on completion. Related to: @conduitFile@.
+--
+-- Since 1.0.9
+conduitHandle :: MonadIO m => IO.Handle -> Conduit S.ByteString m S.ByteString
+conduitHandle h = awaitForever $ \bs -> liftIO (S.hPut h bs) >> yield bs
 
 -- | Ensure that only up to the given number of bytes are consume by the inner
 -- sink. Note that this does /not/ ensure that all of those bytes are in fact
@@ -263,8 +315,8 @@ dropWhile p =
 --
 -- Since 0.3.0
 take :: Monad m => Int -> Consumer S.ByteString m L.ByteString
-take n0 =
-    go n0 id
+take  0 = return L.empty
+take n0 = go n0 id
   where
     go n front =
         await >>= maybe (return $ L.fromChunks $ front []) go'
@@ -281,8 +333,8 @@ take n0 =
 --
 -- Since 0.5.0
 drop :: Monad m => Int -> Consumer S.ByteString m ()
-drop =
-    go
+drop  0 = return ()
+drop n0 = go n0
   where
     go n =
         await >>= maybe (return ()) go'
@@ -359,3 +411,22 @@ sinkCacheLength = do
 -- Since 1.0.5
 sinkLbs :: Monad m => Sink S.ByteString m L.ByteString
 sinkLbs = fmap L.fromChunks consume
+
+mapM_BS :: Monad m => (Word8 -> m ()) -> S.ByteString -> m ()
+mapM_BS f (PS fptr offset len) = do
+    let start = unsafeForeignPtrToPtr fptr `plusPtr` offset
+        end = start `plusPtr` len
+        loop ptr
+            | ptr >= end = inlinePerformIO (touchForeignPtr fptr) `seq` return ()
+            | otherwise = do
+                f (inlinePerformIO (peek ptr))
+                loop (ptr `plusPtr` 1)
+    loop start
+{-# INLINE mapM_BS #-}
+
+-- | Perform a computation on each @Word8@ in a stream.
+--
+-- Since 1.0.10
+mapM_ :: Monad m => (Word8 -> m ()) -> Consumer S.ByteString m ()
+mapM_ f = awaitForever (lift . mapM_BS f)
+{-# INLINE mapM_ #-}
