@@ -1,10 +1,13 @@
 {-# OPTIONS_HADDOCK not-home #-}
 {-# LANGUAGE FlexibleInstances #-}
+{-# LANGUAGE FlexibleContexts #-}
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
 {-# LANGUAGE UndecidableInstances #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE TupleSections #-}
 {-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE ImpredicativeTypes #-}
 module Data.Conduit.Internal
     ( -- * Types
       Pipe (..)
@@ -40,6 +43,13 @@ module Data.Conduit.Internal
     , conduitToPipe
     , toProducer
     , toConsumer
+      -- * Exceptions
+    , catchP
+    , handleP
+    , tryP
+    , catchC
+    , handleC
+    , tryC
       -- * Utilities
     , transPipe
     , mapOutput
@@ -48,10 +58,20 @@ module Data.Conduit.Internal
     , sourceList
     , withUpstream
     , unwrapResumable
+    , Data.Conduit.Internal.enumFromTo
+    , zipSinks
+    , zipSources
+    , zipSourcesApp
     ) where
 
 import Control.Applicative (Applicative (..))
-import Control.Monad ((>=>), liftM, ap, when)
+import Control.Exception.Lifted as E (Exception, catch)
+import Control.Monad ((>=>), liftM, ap, when, liftM2)
+import Control.Monad.Error.Class(MonadError(..))
+import Control.Monad.Reader.Class(MonadReader(..))
+import Control.Monad.RWS.Class(MonadRWS())
+import Control.Monad.Writer.Class(MonadWriter(..))
+import Control.Monad.State.Class(MonadState(..))
 import Control.Monad.Trans.Class (MonadTrans (lift))
 import Control.Monad.IO.Class (MonadIO (liftIO))
 import Control.Monad.Base (MonadBase (liftBase))
@@ -60,7 +80,7 @@ import Data.Monoid (Monoid (mappend, mempty))
 import Control.Monad.Trans.Resource
 import qualified GHC.Exts
 import qualified Data.IORef as I
-import Control.Monad.Morph
+import Control.Monad.Morph (MFunctor (..))
 
 -- | The underlying datatype for all the types in this package.  In has six
 -- type parameters:
@@ -140,6 +160,60 @@ instance Monad m => Monoid (Pipe l i o u m ()) where
 instance MonadResource m => MonadResource (Pipe l i o u m) where
     liftResourceT = lift . liftResourceT
 
+instance MonadReader r m => MonadReader r (Pipe l i o u m) where
+    ask = lift ask
+    local f (HaveOutput p c o) = HaveOutput (local f p) c o
+    local f (NeedInput p c) = NeedInput (\i -> local f (p i)) (\u -> local f (c u))
+    local _ (Done x) = Done x
+    local f (PipeM mp) = PipeM (local f mp)
+    local f (Leftover p i) = Leftover (local f p) i
+
+-- Provided for doctest
+#ifndef MIN_VERSION_mtl
+#define MIN_VERSION_mtl(x, y, z) 0
+#endif
+
+instance MonadWriter w m => MonadWriter w (Pipe l i o u m) where
+#if MIN_VERSION_mtl(2, 1, 0)
+    writer = lift . writer
+#endif
+
+    tell = lift . tell
+
+    listen (HaveOutput p c o) = HaveOutput (listen p) c o
+    listen (NeedInput p c) = NeedInput (\i -> listen (p i)) (\u -> listen (c u))
+    listen (Done x) = Done (x,mempty)
+    listen (PipeM mp) =
+      PipeM $
+      do (p,w) <- listen mp
+         return $ do (x,w') <- listen p
+                     return (x, w `mappend` w')
+    listen (Leftover p i) = Leftover (listen p) i
+
+    pass (HaveOutput p c o) = HaveOutput (pass p) c o
+    pass (NeedInput p c) = NeedInput (\i -> pass (p i)) (\u -> pass (c u))
+    pass (PipeM mp) = PipeM $ mp >>= (return . pass)
+    pass (Done (x,_)) = Done x
+    pass (Leftover p i) = Leftover (pass p) i
+
+instance MonadState s m => MonadState s (Pipe l i o u m) where
+    get = lift get
+    put = lift . put
+#if MIN_VERSION_mtl(2, 1, 0)
+    state = lift . state
+#endif
+
+instance MonadRWS r w s m => MonadRWS r w s (Pipe l i o u m)
+
+instance MonadError e m => MonadError e (Pipe l i o u m) where
+    throwError = lift . throwError
+    catchError (HaveOutput p c o) f = HaveOutput (catchError p f) c o
+    catchError (NeedInput p c) f = NeedInput (\i -> catchError (p i) f) (\u -> catchError (c u) f)
+    catchError (Done x) _ = Done x
+    catchError (PipeM mp) f =
+      PipeM $ catchError (liftM (flip catchError f) mp) (\e -> return (f e))
+    catchError (Leftover p i) f = Leftover (catchError p f) i
+
 -- | Core datatype of the conduit package. This type represents a general
 -- component which can consume a stream of input values @i@, produce a stream
 -- of output values @o@, perform actions in the @m@ monad, and produce a final
@@ -149,8 +223,35 @@ instance MonadResource m => MonadResource (Pipe l i o u m) where
 -- Since 1.0.0
 newtype ConduitM i o m r = ConduitM { unConduitM :: Pipe i i o () m r }
     deriving (Functor, Applicative, Monad, MonadIO, MonadTrans, MonadThrow, MonadActive, MonadResource, MFunctor)
+
+instance MonadReader r m => MonadReader r (ConduitM i o m) where
+    ask = ConduitM ask
+    local f (ConduitM m) = ConduitM (local f m)
+
+instance MonadWriter w m => MonadWriter w (ConduitM i o m) where
+#if MIN_VERSION_mtl(2, 1, 0)
+    writer = ConduitM . writer
+#endif
+    tell = ConduitM . tell
+    listen (ConduitM m) = ConduitM (listen m)
+    pass (ConduitM m) = ConduitM (pass m)
+
+instance MonadState s m => MonadState s (ConduitM i o m) where
+    get = ConduitM get
+    put = ConduitM . put
+#if MIN_VERSION_mtl(2, 1, 0)
+    state = ConduitM . state
+#endif
+
+instance MonadRWS r w s m => MonadRWS r w s (ConduitM i o m)
+
+instance MonadError e m => MonadError e (ConduitM i o m) where
+    throwError = ConduitM . throwError
+    catchError (ConduitM m) f = ConduitM $ catchError m (unConduitM . f)
+
 instance MonadBase base m => MonadBase base (ConduitM i o m) where
     liftBase = lift . liftBase
+
 instance Monad m => Monoid (ConduitM i o m ()) where
     mempty = return ()
     mappend = (>>)
@@ -171,8 +272,10 @@ type Producer m o = forall i. ConduitM i o m ()
 -- | Consumes a stream of input values and produces a final result, without
 -- producing any output.
 --
+-- > type Sink i m r = ConduitM i Void m r
+--
 -- Since 0.5.0
-type Sink i m r = ConduitM i Void m r
+type Sink i = ConduitM i Void
 
 -- | A component which consumes a stream of input values and produces a final
 -- result, regardless of the output stream. A @Consumer@ is a generalization of
@@ -195,6 +298,10 @@ type Conduit i m o = ConduitM i o m ()
 -- Since 0.5.0
 data ResumableSource m o = ResumableSource (Source m o) (m ())
 
+-- | Since 1.0.13
+instance MFunctor ResumableSource where
+    hoist nat (ResumableSource src m) = ResumableSource (hoist nat src) (nat m)
+
 -- | Wait for a single input value from upstream.
 --
 -- Since 0.5.0
@@ -247,8 +354,7 @@ yieldOr o f = HaveOutput (Done ()) f o
 {-# RULES
     "yield o >> p" forall o (p :: Pipe l i o u m r). yield o >> p = HaveOutput p (return ()) o
   ; "mapM_ yield" mapM_ yield = sourceList
-  ; "yieldOr o c >> p" forall o c (p :: Pipe l i o u m r). yieldOr o c >> p = HaveOutput p c o
-  #-}
+  ; "yieldOr o c >> p" forall o c (p :: Pipe l i o u m r). yieldOr o c >> p = HaveOutput p c o #-}
 
 -- | Provide a single piece of leftover input to be consumed by the next pipe
 -- in the current monadic binding.
@@ -466,21 +572,29 @@ transPipe f (Leftover p i) = Leftover (transPipe f p) i
 --
 -- Since 0.4.1
 mapOutput :: Monad m => (o1 -> o2) -> Pipe l i o1 u m r -> Pipe l i o2 u m r
-mapOutput f (HaveOutput p c o) = HaveOutput (mapOutput f p) c (f o)
-mapOutput f (NeedInput p c) = NeedInput (mapOutput f . p) (mapOutput f . c)
-mapOutput _ (Done r) = Done r
-mapOutput f (PipeM mp) = PipeM (liftM (mapOutput f) mp)
-mapOutput f (Leftover p i) = Leftover (mapOutput f p) i
+mapOutput f =
+    go
+  where
+    go (HaveOutput p c o) = HaveOutput (go p) c (f o)
+    go (NeedInput p c) = NeedInput (go . p) (go . c)
+    go (Done r) = Done r
+    go (PipeM mp) = PipeM (liftM (go) mp)
+    go (Leftover p i) = Leftover (go p) i
+{-# INLINE mapOutput #-}
 
 -- | Same as 'mapOutput', but use a function that returns @Maybe@ values.
 --
 -- Since 0.5.0
 mapOutputMaybe :: Monad m => (o1 -> Maybe o2) -> Pipe l i o1 u m r -> Pipe l i o2 u m r
-mapOutputMaybe f (HaveOutput p c o) = maybe id (\o' p' -> HaveOutput p' c o') (f o) (mapOutputMaybe f p)
-mapOutputMaybe f (NeedInput p c) = NeedInput (mapOutputMaybe f . p) (mapOutputMaybe f . c)
-mapOutputMaybe _ (Done r) = Done r
-mapOutputMaybe f (PipeM mp) = PipeM (liftM (mapOutputMaybe f) mp)
-mapOutputMaybe f (Leftover p i) = Leftover (mapOutputMaybe f p) i
+mapOutputMaybe f =
+    go
+  where
+    go (HaveOutput p c o) = maybe id (\o' p' -> HaveOutput p' c o') (f o) (mapOutputMaybe f p)
+    go (NeedInput p c) = NeedInput (go . p) (go . c)
+    go (Done r) = Done r
+    go (PipeM mp) = PipeM (liftM (go) mp)
+    go (Leftover p i) = Leftover (go p) i
+{-# INLINE mapOutputMaybe #-}
 
 -- | Apply a function to all the input values of a @Pipe@.
 --
@@ -496,6 +610,18 @@ mapInput _ _  (Done r)           = Done r
 mapInput f f' (PipeM mp)         = PipeM (liftM (mapInput f f') mp)
 mapInput f f' (Leftover p i)     = maybe id (flip Leftover) (f' i) $ mapInput f f' p
 
+enumFromTo :: (Enum o, Eq o, Monad m)
+           => o
+           -> o
+           -> Pipe l i o u m ()
+enumFromTo start stop =
+    loop start
+  where
+    loop i
+        | i == stop = HaveOutput (Done ()) (return ()) i
+        | otherwise = HaveOutput (loop (succ i)) (return ()) i
+{-# INLINE enumFromTo #-}
+
 -- | Convert a list into a source.
 --
 -- Since 0.3.0
@@ -514,8 +640,7 @@ build :: Monad m => (forall b. (o -> b -> b) -> b -> b) -> Pipe l i o u m ()
 build g = g (\o p -> HaveOutput p (return ()) o) (return ())
 
 {-# RULES
-    "sourceList/build" forall (f :: (forall b. (a -> b -> b) -> b -> b)). sourceList (GHC.Exts.build f) = build f
-  #-}
+    "sourceList/build" forall (f :: (forall b. (a -> b -> b) -> b -> b)). sourceList (GHC.Exts.build f) = build f #-}
 
 sourceToPipe :: Monad m => Source m o -> Pipe l i o u m ()
 sourceToPipe =
@@ -636,3 +761,155 @@ toConsumer =
 -- | Since 1.0.4
 instance MFunctor (Pipe l i o u) where
     hoist = transPipe
+
+-- | See 'catchC' for more details.
+--
+-- Since 1.0.11
+catchP :: (MonadBaseControl IO m, Exception e)
+       => Pipe l i o u m r
+       -> (e -> Pipe l i o u m r)
+       -> Pipe l i o u m r
+catchP p0 onErr =
+    go p0
+  where
+    go (Done r) = Done r
+    go (PipeM mp) = PipeM $ E.catch (liftM go mp) (return . onErr)
+    go (Leftover p i) = Leftover (go p) i
+    go (NeedInput x y) = NeedInput (go . x) (go . y)
+    go (HaveOutput p c o) = HaveOutput (go p) c o
+{-# INLINABLE catchP #-}
+
+-- | The same as @flip catchP@.
+--
+-- Since 1.0.11
+handleP :: (MonadBaseControl IO m, Exception e)
+        => (e -> Pipe l i o u m r)
+        -> Pipe l i o u m r
+        -> Pipe l i o u m r
+handleP = flip catchP
+{-# INLINE handleP #-}
+
+-- | See 'tryC' for more details.
+--
+-- Since 1.0.11
+tryP :: (MonadBaseControl IO m, Exception e)
+     => Pipe l i o u m r
+     -> Pipe l i o u m (Either e r)
+tryP =
+    go
+  where
+    go (Done r) = Done (Right r)
+    go (PipeM mp) = PipeM $ E.catch (liftM go mp) (return . Done . Left)
+    go (Leftover p i) = Leftover (go p) i
+    go (NeedInput x y) = NeedInput (go . x) (go . y)
+    go (HaveOutput p c o) = HaveOutput (go p) c o
+{-# INLINABLE tryP #-}
+
+-- | Catch all exceptions thrown by the current component of the pipeline.
+--
+-- Note: this will /not/ catch exceptions thrown by other components! For
+-- example, if an exception is thrown in a @Source@ feeding to a @Sink@, and
+-- the @Sink@ uses @catchC@, the exception will /not/ be caught.
+--
+-- Due to this behavior (as well as lack of async exception handling), you
+-- should not try to implement combinators such as @onException@ in terms of this
+-- primitive function.
+--
+-- Note also that the exception handling will /not/ be applied to any
+-- finalizers generated by this conduit.
+--
+-- Since 1.0.11
+catchC :: (MonadBaseControl IO m, Exception e)
+       => ConduitM i o m r
+       -> (e -> ConduitM i o m r)
+       -> ConduitM i o m r
+catchC (ConduitM p) f = ConduitM (catchP p (unConduitM . f))
+{-# INLINE catchC #-}
+
+-- | The same as @flip catchC@.
+--
+-- Since 1.0.11
+handleC :: (MonadBaseControl IO m, Exception e)
+        => (e -> ConduitM i o m r)
+        -> ConduitM i o m r
+        -> ConduitM i o m r
+handleC = flip catchC
+{-# INLINE handleC #-}
+
+-- | A version of @try@ for use within a pipeline. See the comments in @catchC@
+-- for more details.
+--
+-- Since 1.0.11
+tryC :: (MonadBaseControl IO m, Exception e)
+     => ConduitM i o m r
+     -> ConduitM i o m (Either e r)
+tryC = ConduitM . tryP . unConduitM
+{-# INLINE tryC #-}
+
+-- | Combines two sinks. The new sink will complete when both input sinks have
+--   completed.
+--
+-- Any leftovers are discarded.
+--
+-- Since 0.4.1
+zipSinks :: Monad m => Sink i m r -> Sink i m r' -> Sink i m (r, r')
+zipSinks (ConduitM x0) (ConduitM y0) =
+    ConduitM $ injectLeftovers x0 >< injectLeftovers y0
+  where
+    (><) :: Monad m => Pipe Void i Void () m r1 -> Pipe Void i Void () m r2 -> Pipe l i o () m (r1, r2)
+
+    Leftover _  i    >< _                = absurd i
+    _                >< Leftover _  i    = absurd i
+    HaveOutput _ _ o >< _                = absurd o
+    _                >< HaveOutput _ _ o = absurd o
+
+    PipeM mx         >< y                = PipeM (liftM (>< y) mx)
+    x                >< PipeM my         = PipeM (liftM (x ><) my)
+    Done x           >< Done y           = Done (x, y)
+    NeedInput px cx  >< NeedInput py cy  = NeedInput (\i -> px i >< py i) (\() -> cx () >< cy ())
+    NeedInput px cx  >< y@Done{}         = NeedInput (\i -> px i >< y)    (\u -> cx u >< y)
+    x@Done{}         >< NeedInput py cy  = NeedInput (\i -> x >< py i)    (\u -> x >< cy u)
+
+-- | Combines two sources. The new source will stop producing once either
+--   source has been exhausted.
+--
+-- Since 1.0.13
+zipSources :: Monad m => Source m a -> Source m b -> Source m (a, b)
+zipSources (ConduitM left0) (ConduitM right0) =
+    ConduitM $ go left0 right0
+  where
+    go (Leftover left ()) right = go left right
+    go left (Leftover right ())  = go left right
+    go (Done ()) (Done ()) = Done ()
+    go (Done ()) (HaveOutput _ close _) = PipeM (close >> return (Done ()))
+    go (HaveOutput _ close _) (Done ()) = PipeM (close >> return (Done ()))
+    go (Done ()) (PipeM _) = Done ()
+    go (PipeM _) (Done ()) = Done ()
+    go (PipeM mx) (PipeM my) = PipeM (liftM2 go mx my)
+    go (PipeM mx) y@HaveOutput{} = PipeM (liftM (\x -> go x y) mx)
+    go x@HaveOutput{} (PipeM my) = PipeM (liftM (go x) my)
+    go (HaveOutput srcx closex x) (HaveOutput srcy closey y) = HaveOutput (go srcx srcy) (closex >> closey) (x, y)
+    go (NeedInput _ c) right = go (c ()) right
+    go left (NeedInput _ c) = go left (c ())
+
+-- | Combines two sources. The new source will stop producing once either
+--   source has been exhausted.
+--
+-- Since 1.0.13
+zipSourcesApp :: Monad m => Source m (a -> b) -> Source m a -> Source m b
+zipSourcesApp (ConduitM left0) (ConduitM right0) =
+    ConduitM $ go left0 right0
+  where
+    go (Leftover left ()) right = go left right
+    go left (Leftover right ())  = go left right
+    go (Done ()) (Done ()) = Done ()
+    go (Done ()) (HaveOutput _ close _) = PipeM (close >> return (Done ()))
+    go (HaveOutput _ close _) (Done ()) = PipeM (close >> return (Done ()))
+    go (Done ()) (PipeM _) = Done ()
+    go (PipeM _) (Done ()) = Done ()
+    go (PipeM mx) (PipeM my) = PipeM (liftM2 go mx my)
+    go (PipeM mx) y@HaveOutput{} = PipeM (liftM (\x -> go x y) mx)
+    go x@HaveOutput{} (PipeM my) = PipeM (liftM (go x) my)
+    go (HaveOutput srcx closex x) (HaveOutput srcy closey y) = HaveOutput (go srcx srcy) (closex >> closey) (x y)
+    go (NeedInput _ c) right = go (c ()) right
+    go left (NeedInput _ c) = go left (c ())
