@@ -23,10 +23,16 @@ module Data.Conduit.Text
     , lines
     , linesBounded
     , TextException (..)
+    , takeWhile
+    , dropWhile
+    , take
+    , drop
+    , foldLines
+    , withLine
     ) where
 
 import qualified Prelude
-import           Prelude hiding (head, drop, takeWhile, lines, zip, zip3, zipWith, zipWith3)
+import           Prelude hiding (head, drop, takeWhile, lines, zip, zip3, zipWith, zipWith3, take, dropWhile)
 
 import           Control.Arrow (first)
 import qualified Control.Exception as Exc
@@ -204,32 +210,34 @@ utf8 :: Codec
 utf8 = Codec name enc dec where
     name = T.pack "UTF-8"
     enc text = (TE.encodeUtf8 text, Nothing)
-    dec bytes = case splitQuickly bytes of
+    dec bytes = case splitQuickly bytes >>= maybeDecode of
         Just (text, extra) -> (text, Right extra)
         Nothing -> splitSlowly TE.decodeUtf8 bytes
 
-    splitQuickly bytes = loop 0 >>= maybeDecode where
-        required x0
-            | x0 .&. 0x80 == 0x00 = 1
-            | x0 .&. 0xE0 == 0xC0 = 2
-            | x0 .&. 0xF0 == 0xE0 = 3
-            | x0 .&. 0xF8 == 0xF0 = 4
+    -- Whether the given byte is a continuation byte.
+    isContinuation byte = byte .&. 0xC0 == 0x80
 
-            -- Invalid input; let Text figure it out
-            | otherwise           = 0
+    -- The number of continuation bytes needed by the given
+    -- non-continuation byte. Returns -1 for an illegal UTF-8
+    -- non-continuation byte and the whole split quickly must fail so
+    -- as the input is passed to TE.decodeUtf8, which will issue a
+    -- suitable error.
+    required x0
+        | x0 .&. 0x80 == 0x00 = 0
+        | x0 .&. 0xE0 == 0xC0 = 1
+        | x0 .&. 0xF0 == 0xE0 = 2
+        | x0 .&. 0xF8 == 0xF0 = 3
+        | otherwise           = -1
 
-        maxN = B.length bytes
-
-        loop n | n == maxN = Just (TE.decodeUtf8 bytes, B.empty)
-        loop n = let
-            req = required (B.index bytes n)
-            tooLong = first TE.decodeUtf8 (B.splitAt n bytes)
-            decodeMore = loop $! n + req
-            in if req == 0
-                then Nothing
-                else if n + req > maxN
-                    then Just tooLong
-                    else decodeMore
+    splitQuickly bytes
+        | B.null l || req == -1 = Nothing
+        | req == B.length r = Just (TE.decodeUtf8 bytes, B.empty)
+        | otherwise = Just (TE.decodeUtf8 l', r')
+      where
+        (l, r) = B.spanEnd isContinuation bytes
+        req = required (B.last l)
+        l' = B.init l
+        r' = B.cons (B.last l) r
 
 -- |
 -- Since 0.3.0
@@ -366,3 +374,105 @@ maybeDecode :: (a, b) -> Maybe (a, b)
 maybeDecode (a, b) = case tryEvaluate a of
     Left _ -> Nothing
     Right _ -> Just (a, b)
+
+-- |
+--
+-- Since 1.0.8
+takeWhile :: Monad m
+          => (Char -> Bool)
+          -> Conduit T.Text m T.Text
+takeWhile p =
+    loop
+  where
+    loop = await >>= maybe (return ()) go
+    go t =
+        case T.span p t of
+            (x, y)
+                | T.null y -> yield x >> loop
+                | otherwise -> yield x >> leftover y
+
+-- |
+--
+-- Since 1.0.8
+dropWhile :: Monad m
+          => (Char -> Bool)
+          -> Consumer T.Text m ()
+dropWhile p =
+    loop
+  where
+    loop = await >>= maybe (return ()) go
+    go t
+        | T.null x = loop
+        | otherwise = leftover x
+      where
+        x = T.dropWhile p t
+
+-- |
+--
+-- Since 1.0.8
+take :: Monad m => Int -> Conduit T.Text m T.Text
+take =
+    loop
+  where
+    loop i = await >>= maybe (return ()) (go i)
+    go i t
+        | diff == 0 = yield t
+        | diff < 0 =
+            let (x, y) = T.splitAt i t
+             in yield x >> leftover y
+        | otherwise = yield t >> loop diff
+      where
+        diff = i - T.length t
+
+-- |
+--
+-- Since 1.0.8
+drop :: Monad m => Int -> Consumer T.Text m ()
+drop =
+    loop
+  where
+    loop i = await >>= maybe (return ()) (go i)
+    go i t
+        | diff == 0 = return ()
+        | diff < 0 = leftover $ T.drop i t
+        | otherwise = loop diff
+      where
+        diff = i - T.length t
+
+-- |
+--
+-- Since 1.0.8
+foldLines :: Monad m
+          => (a -> ConduitM T.Text o m a)
+          -> a
+          -> ConduitM T.Text o m a
+foldLines f =
+    start
+  where
+    start a = CL.peek >>= maybe (return a) (const $ loop $ f a)
+
+    loop consumer = do
+        a <- takeWhile (/= '\n') =$= do
+            a <- CL.map (T.filter (/= '\r')) =$= consumer
+            CL.sinkNull
+            return a
+        drop 1
+        start a
+
+-- |
+--
+-- Since 1.0.8
+withLine :: Monad m
+         => Sink T.Text m a
+         -> Consumer T.Text m (Maybe a)
+withLine consumer = toConsumer $ do
+    mx <- CL.peek
+    case mx of
+        Nothing -> return Nothing
+        Just _ -> do
+            x <- takeWhile (/= '\n') =$ do
+                x <- CL.map (T.filter (/= '\r')) =$ consumer
+                CL.sinkNull
+                return x
+            drop 1
+            return $ Just x
