@@ -4,7 +4,9 @@ import Test.Hspec
 import Test.Hspec.QuickCheck (prop)
 import Test.QuickCheck.Monadic (assert, monadicIO, run)
 
+import Control.Exception (IOException)
 import qualified Data.Conduit as C
+import qualified Data.Conduit.Lift as C
 import qualified Data.Conduit.Util as C
 import qualified Data.Conduit.Internal as CI
 import qualified Data.Conduit.List as CL
@@ -26,13 +28,14 @@ import qualified Data.Text.Lazy as TL
 import qualified Data.Text.Lazy.Encoding as TLE
 import Control.Monad.Trans.Resource (runExceptionT, runExceptionT_, allocate, resourceForkIO)
 import Control.Concurrent (threadDelay, killThread)
-import Control.Monad.IO.Class (liftIO)
+import Control.Monad.IO.Class (MonadIO, liftIO)
 import Control.Monad.Trans.Class (lift)
 import Control.Monad.Trans.Writer (execWriter, tell, runWriterT)
-import Control.Monad.Trans.State (evalStateT, get, put)
+import Control.Monad.Trans.State (evalStateT, get, put, modify)
+import Control.Monad.Trans.Maybe (MaybeT (..))
 import Control.Applicative (pure, (<$>), (<*>))
 import Data.Functor.Identity (Identity,runIdentity)
-import Control.Monad (forever)
+import Control.Monad (forever, void)
 import Data.Void (Void)
 import qualified Control.Concurrent.MVar as M
 import Control.Monad.Error (catchError, throwError, Error)
@@ -191,7 +194,7 @@ main = hspec $ do
 
     describe "zipping" $ do
         it "zipping two small lists" $ do
-            res <- runResourceT $ C.zip (CL.sourceList [1..10]) (CL.sourceList [11..12]) C.$$ CL.consume
+            res <- runResourceT $ C.zipSources (CL.sourceList [1..10]) (CL.sourceList [11..12]) C.$$ CL.consume
             res @=? zip [1..10 :: Int] [11..12 :: Int]
 
     describe "zipping sinks" $ do
@@ -936,6 +939,33 @@ main = hspec $ do
                 lbs' `shouldBe` lbs
                 fromIntegral len `shouldBe` L.length lbs'
 
+    describe "Data.Conduit.Binary.mapM_" $ do
+        prop "telling works" $ \bytes ->
+            let lbs = L.pack bytes
+                src = CB.sourceLbs lbs
+                sink = CB.mapM_ (tell . return . S.singleton)
+                bss = execWriter $ src C.$$ sink
+             in L.fromChunks bss == lbs
+
+    describe "passthroughSink" $ do
+        it "works" $ do
+            ref <- I.newIORef (-1)
+            let sink = CL.fold (+) (0 :: Int)
+                conduit = C.passthroughSink sink (I.writeIORef ref)
+                input = [1..10]
+            output <- mapM_ C.yield input C.$$ conduit C.=$ CL.consume
+            output `shouldBe` input
+            x <- I.readIORef ref
+            x `shouldBe` sum input
+        it "does nothing when downstream does nothing" $ do
+            ref <- I.newIORef (-1)
+            let sink = CL.fold (+) (0 :: Int)
+                conduit = C.passthroughSink sink (I.writeIORef ref)
+                input = [undefined]
+            mapM_ C.yield input C.$$ conduit C.=$ return ()
+            x <- I.readIORef ref
+            x `shouldBe` (-1)
+
     describe "mtl instances" $ do
         it "ErrorT" $ do
             let src = flip catchError (const $ C.yield 4) $ do
@@ -1049,6 +1079,63 @@ main = hspec $ do
             -- FIXME this is broken it "test2" $ verify test2L test2R "p2" "p1" "p3"
             it "test3" $ verify test3L test3R "p2" "p3" "p1"
 
+    describe "Data.Conduit.Lift" $ do
+        it "execStateC" $ do
+            let sink = C.execStateC 0 $ CL.mapM_ $ modify . (+)
+                src = mapM_ C.yield [1..10 :: Int]
+            res <- src C.$$ sink
+            res `shouldBe` sum [1..10]
+
+        it "execWriterC" $ do
+            let sink = C.execWriterC $ CL.mapM_ $ tell . return
+                src = mapM_ C.yield [1..10 :: Int]
+            res <- src C.$$ sink
+            res `shouldBe` [1..10]
+
+        it "runErrorC" $ do
+            let sink = C.runErrorC $ do
+                    x <- C.catchErrorC (lift $ throwError "foo") return
+                    return $ x ++ "bar"
+            res <- return () C.$$ sink
+            res `shouldBe` Right ("foobar" :: String)
+
+        it "runMaybeC" $ do
+            let src = void $ C.runMaybeC $ do
+                    C.yield 1
+                    () <- lift $ MaybeT $ return Nothing
+                    C.yield 2
+                sink = CL.consume
+            res <- src C.$$ sink
+            res `shouldBe` [1 :: Int]
+
+    describe "exception handling" $ do
+        it "catchC" $ do
+            ref <- I.newIORef 0
+            let src = do
+                    C.catchC (CB.sourceFile "some-file-that-does-not-exist") onErr
+                    C.handleC onErr $ CB.sourceFile "conduit.cabal"
+                onErr :: MonadIO m => IOException -> m ()
+                onErr _ = liftIO $ I.modifyIORef ref (+ 1)
+            contents <- L.readFile "conduit.cabal"
+            res <- C.runResourceT $ src C.$$ CB.sinkLbs
+            res `shouldBe` contents
+            errCount <- I.readIORef ref
+            errCount `shouldBe` (1 :: Int)
+        it "tryC" $ do
+            ref <- I.newIORef undefined
+            let src = do
+                    res1 <- C.tryC $ CB.sourceFile "some-file-that-does-not-exist"
+                    res2 <- C.tryC $ CB.sourceFile "conduit.cabal"
+                    liftIO $ I.writeIORef ref (res1, res2)
+            contents <- L.readFile "conduit.cabal"
+            res <- C.runResourceT $ src C.$$ CB.sinkLbs
+            res `shouldBe` contents
+            exc <- I.readIORef ref
+            case exc :: (Either IOException (), Either IOException ()) of
+                (Left _, Right ()) ->
+                    return ()
+                _ -> error $ show exc
+
 it' :: String -> IO () -> Spec
 it' = it
 
